/*
  Power Meter

  T. Nakagawa

  2010-03-21
*/


#include <math.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include "uart.h"
#include "suart.h"
#include "14seglcd.h"


#define OSCCALVAL 182
#define REFVLT	2.731	/* ADC̊d(AVR 21ԃs̓d) */
#define LOGAMP_DBMVLT	40.0	/* OAvo͂dBm/volt */
#define LOGAMP_OFST	(-90.0 + 1.86)	/* OAvo͂̃ItZbg */
#define FRMS_VLTMAX 2.50	/* FRMS[hPC֑Ml256̏ꍇɉVɑ邩 */

#define MODE_DBM 0
#define MODE_WATT 1
#define MODE_VRMS 2
#define MODE_FRMS 3


volatile static uint8_t Mode;
volatile static uint32_t Frms_start, Frms_inc, Frms_wait;


/*
  L[͂̏([hؑ)
*/
static int8_t keysense(void) {
  static uint8_t cnt = 0;
  static uint8_t prv = 0xff;
  uint8_t tmp;

  if (++cnt != 12) return 0;	/* 1250HzŌĂ΂̂Ŗ100HzɃL[̃TvO */
  cnt = 0;

  tmp = prv;
  prv = PINB;
  if (((tmp ^ prv) & ~prv & _BV(PD0)) != 0) {	/* XCb`ꂽ */
    Mode++;
    Mode &= 0x03;

    if (Mode == MODE_FRMS) {
      UBRRL = (uint8_t)((double)F_CPU / (8.0 * 19200) - 1.0 + 0.5);
    } else {
      UBRRL = (uint8_t)((double)F_CPU / (8.0 * 115200) - 1.0 + 0.5);
    }

    return 1;
  }

  return 0;
}


/*
  d͂̑(AD8307̏o͓d*1024/2.56VԂ)
  130(129.6))kHz/13/8=1250Hz̃TvO
  FRMSłADCɂ鎞Ԃ1216HzƂȂĂ
  L[͂`^O邽߂̏ɍs
*/
static uint16_t measure(void) {
  uint8_t i;
  uint16_t tmp, adc;

  adc = 0;
  for (i = 0; i < 8; i++) {
    /* ADϊ(mCYጸ) */
    sei();
    sleep_mode();
    cli();
    tmp = ADCW;
    adc += tmp;
  }
  adc /= 8;

  keysense();

  return adc;
}


/*
  [hɉēd́^dLCDɕ\ & UARTɏo or bZ[W\
*/
static void display(uint16_t adc) {
  uint8_t i, j;
  double dbm, watt, vrms;
  uint8_t pnt;
  uint16_t num;

  if (Mode == MODE_FRMS) {
    lcd_buf[0] = LCD_SPACE;
    lcd_buf[1] = LCD_SPACE;
    lcd_buf[2] = LCD_F;
    lcd_buf[3] = LCD_R;
    lcd_buf[4] = LCD_M;
    lcd_buf[5] = LCD_S;
    lcd_buf[6] = LCD_SPACE;
    lcd_buf[7] = LCD_SPACE;
    lcd_buf[8] = 0x00;
  } else {
    dbm = adc * LOGAMP_DBMVLT * REFVLT / 1024.0 + LOGAMP_OFST;
    watt = pow(10.0, dbm / 10.0 + 3.0);	/* watt̒PʂuWƂĂ */
    vrms = sqrt(50.0 * watt);	/* vrms̒PʂmVƂĂ */

    lcd_buf[0] = LCD_SPACE;	/* sign */
    if (Mode == MODE_DBM) {
      lcd_buf[5] = LCD_d;
      lcd_buf[6] = LCD_B;
      lcd_buf[7] = LCD_m;
      lcd_buf[8] = 0x40;
      i = 4;
      if (dbm < 0.0) {
        lcd_buf[0] = LCD_MINUS;
        dbm = -dbm;
      }
      if (dbm >= 10.0) {
        pnt = 3;
        dbm *= 10.0;
      } else if (dbm >= 1.0) {
        pnt = 2;
        dbm *= 100.0;
      } else {
        pnt = 1;
        dbm *= 1000.0;
      }
      num = (uint16_t)dbm;
    } else if (Mode == MODE_WATT) {
      lcd_buf[7] = LCD_W;
      lcd_buf[8] = 0x20;
      i = 5;
      if (watt >= 0.1) {
        lcd_buf[6] = LCD_u;
        if (watt >= 10000.0) {
          pnt = 6;
        } else if (watt >= 1000.0) {
          pnt = 5;
        } else if (watt >= 100.0) {
          pnt = 4;
          watt *= 10.0;
        } else if (watt >= 10.0) {
          pnt = 3;
          watt *= 100.0;
        } else if (watt >= 1.0) {
          pnt = 2;
          watt *= 1000.0;
        } else {
          pnt = 1;
          watt *= 10000.0;
        }
      } else {
        lcd_buf[6] = LCD_n;
        if (watt >= 0.01) {
          pnt = 3;
          watt *= 100000.0;
        } else if (watt >= 0.001) {
          pnt = 2;
          watt *= 1000000.0;
        } else {
          pnt = 1;
          watt *= 10000000.0;
        }
      }
      num = (uint16_t)watt;
    } else if (Mode == MODE_VRMS) {
      lcd_buf[6] = LCD_m;
      lcd_buf[7] = LCD_V;
      lcd_buf[8] = 0x10;
      i = 5;
      if (vrms >= 1000.0) {
        pnt = 5;
      } else if (vrms >= 100.0) {
        pnt = 4;
        vrms *= 10;
      } else if (vrms >= 10.0) {
        pnt = 3;
        vrms *= 100;
      } else if (vrms >= 1.0) {
        pnt = 2;
        vrms *= 1000;
      } else {
        pnt = 1;
        vrms *= 10000;
      }
      num = (uint16_t)vrms;
    }

    /* LCDo */
    j = i;
    for (; i >= 1; i--) {
      if (i == pnt) {
        lcd_buf[i] = LCD_PERIOD;
      } else {
        uint8_t d;
        d = (uint8_t)(num % 10);
        lcd_buf[i] = d;
        num /= 10;
      }
    }
    /* UARTo */
    for (; i <= j; i++) {
      uint8_t c;

      c = lcd_buf[i];
      if (c == LCD_SPACE) uart_putchar(' ');
      else if (c == LCD_MINUS) uart_putchar('-');
      else if (c == LCD_PERIOD) uart_putchar('.');
      else uart_putchar('0' + c);
    }
    uart_putchar('\r');
    uart_putchar('\n');
  }
  lcd_refresh();

  return;
}


/*
  16iASCII(e.g. "FE1234AB")oCilɕϊ
*/
static uint32_t ascii2bin(uint8_t *str, uint8_t bgn, uint8_t end) {
  uint8_t i;
  uint32_t val;
  uint8_t c;

  val = 0;
  for (i = bgn; i < end; i++) {
    val *= 16;
    c = str[i];
    if (c <= '9') {
      val += c - '0';
    } else {
      val += c - 'A' + 10;
    }
  }

  return val;
}


/*
  FRMS[hłPC̃f[^M
*/
static void frms_receive(void) {
  uint8_t i;
  uint8_t c;
  uint8_t buf[18];
  uint16_t cnt = 0;

  /* PCf[^M & L[ZX */
  for (i = 0; ; i++) {
    /* L[ */
    while (bit_is_clear(UCSRA, RXC)) {
      if (++cnt == 1000) {	/* K1250Hz炢ŃTvO悤 */
        if (keysense()) return;
        cnt = 0;
      }
    }
    c = UDR;

    if (c == '\x0d') {
      if (i == 18) break;
      i = 0;
      continue;
    }
    if (i == 18) i = 0;
    buf[i] = c;
  }

  /* XC[vJng */
  Frms_start = ascii2bin(buf, 0, 8);
  /* XC[vg */
  Frms_inc = ascii2bin(buf, 8, 16);
  /* XC[vEFCg */
  Frms_wait = ascii2bin(buf, 16, 18);

  return;
}


/*
  FRMS[hDDS̎gݒ
  32bitʃoCg瑗M
*/
static void dds_send(uint32_t frq) {
  uint8_t d0, d1, d2, d3;

  d3 = (uint8_t)(frq & 0x000000ff);
  frq >>= 8;
  d2 = (uint8_t)(frq & 0x000000ff);
  frq >>= 8;
  d1 = (uint8_t)(frq & 0x000000ff);
  frq >>= 8;
  d0 = (uint8_t)frq;

  suart_putchar(d0);
  suart_putchar(d1);
  suart_putchar(d2);
  suart_putchar(d3);

  return;
}


/*
  FRMS[hł̃EFCg
  411usPʂƂFrms_wait̐EFCg
*/
static void wait(void) {
  uint8_t i;

  for (i = 0; i < Frms_wait; i++) {
    _delay_us(411);
  }

  return;
}


/*
  FRMS[hłPCւ̃f[^M
  l̕␳đM
  Ml=(adc / 1024 * REFVLT) * (256 / FRMS_VLTMAX)
  REFVLT/FRMS_VLTMAX<2ƉDŒ菬_ZD
*/
static void frms_send(uint16_t adc) {
  uint32_t data;
  uint8_t d0, d1;

  data = adc;
  data *= (uint32_t)(REFVLT / FRMS_VLTMAX * 65536.0 * 32.0);	/* X(1+16+5)bitȂ */
  data >>= 23;
  d1 = (uint8_t)(data & 0x0000000f);
  data >>= 4;
  d0 = (uint8_t)(data & 0x0000000f);

  if (d0 <= 9) {
    uart_putchar('0' + d0);
  } else {
    uart_putchar('A' + d0 - 10);
  }
  if (d1 <= 9) {
    uart_putchar('0' + d1);
  } else {
    uart_putchar('A' + d1 - 10);
  }

  return;
}


int main(void) {
  /* NbN̏C(8,294,400Hz) */
  OSCCAL = OSCCALVAL;

  /* |[g */
  PORTB |= (_BV(PB0) | _BV(PB3) | _BV(PB4) | _BV(PB5));

  /* LCD */
  lcd_init();

  /* UART */
  uart_init();

  /* Software UART */
  suart_init();

  /* AiORp[^~ */
  ACSR |= _BV(ACD);

  /* ADC */
  ADMUX |= (_BV(REFS1) | _BV(REFS0));
  ADCSRA |= (_BV(ADEN) | _BV(ADIE) | _BV(ADPS2) | _BV(ADPS1));	/* ADC clock=CK/64=129.6kHz */
  set_sleep_mode(SLEEP_MODE_ADC);

  /*  */
  Mode = MODE_DBM;

  for (; ; ) {	/* X^hA[[h */
    if (Mode != MODE_FRMS) {
      uint16_t i;
      uint32_t val;

      /* v */
      val = 0;
      for (i = 0; i < 512; i++) {
        val += measure();
        if (Mode == MODE_FRMS) break;
      }
      val /= 512;

      /* LCD\ & UARTo */
      display((uint16_t)val);
    } else {	/* FRMS[h */
      uint8_t i;
      uint32_t f;
      uint16_t val;

      frms_receive();
      if (Mode != MODE_FRMS) continue;

      i = 0;
      f = Frms_start;
      do {
        /* DDS̐ݒ */
        dds_send(f);
        f += Frms_inc;
        /* EFCg */
        wait();
        /* v */
        val = measure();
        /* PC֑M */
        frms_send(val);
      } while (++i != 0);
    }
  }

  return 0;
}


/*
  ADC荞(sleep畜A邾)
*/
ISR(ADC_vect) {
}
