/*
  Half Duplex Interrupt Driven Software UART (AVR305)
*/


#include <avr/io.h>


#define BAUD 115200
#define SB 1	/* Number of stop bits (1, 2, ...) */
#define TXENB 1
#define RXENB 0
#if TXENB
#define TX_DDR DDRD	/* Transmit port */
#define TX_PORT PORTD	/* Transmit port */
#define TX_BIT PD2	/* Transmit pin */
#endif
#if RXENB
#define RX_DDR DDRD	/* Receive port */
#define RX_PORT PORTD	/* Receive port */
#define RX_BIT PIND2	/* Receive pin */
#endif

#define B ((F_CPU - 23 * BAUD) / (6 * BAUD))	/* actual baud rate=F_CPU/(6 * B + 23) */
#define bitcnt r23	/* bit counter */
#define temp r22	/* temporary storage register */
#define txbyte r24	/* Data to be transmitted */
#define rxbyte r24	/* Received data */


;; uart_delay
uart_delay:
	ldi	temp, B
1:	dec	temp
	brne	1b
	ret


;; suart_init
.global suart_init
.func suart_init
suart_init:
#if TXENB
	sbi	_SFR_IO_ADDR(TX_DDR), TX_BIT
#endif
#if RXENB
	cbi	_SFR_IO_ADDR(RX_DDR), RX_BIT
#endif

	ret
.endfunc


;; suart_putchar
#if TXENB
.global suart_putchar
.func suart_putchar
suart_putchar:
	ldi	bitcnt, 9 + SB
	com	txbyte
	sec
1:	brcc	2f
	cbi	_SFR_IO_ADDR(TX_PORT), TX_BIT
	rjmp	3f
2:	sbi	_SFR_IO_ADDR(TX_PORT), TX_BIT
	nop
3:	rcall	uart_delay
	rcall	uart_delay
	lsr	txbyte
	dec	bitcnt
	brne	1b
	ret
.endfunc
#endif

;; suart_getchar
#if RXENB
.global suart_getchar
.func suart_getchar
suart_getchar:
	ldi	bitcnt, 9
1:	sbic	_SFR_IO_ADDR(RX_PORT), RX_BIT
	rjmp	1b
	rcall	uart_delay
2:	rcall	uart_delay
	rcall	uart_delay
	clc
	sbic	_SFR_IO_ADDR(RX_PORT), RX_BIT
	sec
	dec	bitcnt
	breq	3f
	ror	rxbyte
	rjmp	2b
3:	ret
.endfunc
#endif
